"""
Estimate nonlinear normalization parameters from Freesurfer conformed space
to FSL's nonlinear MNI152 target using FLIRT and FNIRT.

See the docstring for the create_normalization_workflow function for more
information about the processing.

"""
import os
import os.path as op
from moss.mosaic import Mosaic

from nipype import (IdentityInterface,
                    SelectFiles, DataSink,
                    Node, Workflow)
from nipype.interfaces import fsl, freesurfer as fs
from nipype.interfaces.base import (BaseInterface,
                                    BaseInterfaceInputSpec,
                                    TraitedSpec, File)

from lyman.tools import submit_cmdline


def create_fsl_workflow(data_dir=None, subjects=None, name="fslwarp"):
    """Set up the anatomical normalzation workflow using FNIRT.

    Your anatomical data must have been processed in Freesurfer.
    Unlike most lyman workflows, the DataGrabber and DataSink
    nodes are hardwired within the returned workflow, as this
    tightly integrates with the Freesurfer subjects directory
    structure.

    Parameters
    ----------
    data_dir : path
        top level of data hierarchy/FS subjects directory
    subjects : list of strings
        list of subject IDs
    name : alphanumeric string, optional
        workflow name

    """
    if data_dir is None:
        data_dir = os.environ["SUBJECTS_DIR"]
    if subjects is None:
        subjects = []

    # Get target images
    target_brain = fsl.Info.standard_image("avg152T1_brain.nii.gz")
    target_head = fsl.Info.standard_image("avg152T1.nii.gz")
    hires_head = fsl.Info.standard_image("MNI152_T1_1mm.nii.gz")
    target_mask = fsl.Info.standard_image(
        "MNI152_T1_2mm_brain_mask_dil.nii.gz")
    fnirt_cfg = os.path.join(
        os.environ["FSLDIR"], "etc/flirtsch/T1_2_MNI152_2mm.cnf")

    # Subject source node
    subjectsource = Node(IdentityInterface(fields=["subject_id"]),
                         iterables=("subject_id", subjects),
                         name="subjectsource")

    # Grab recon-all outputs
    head_image = "T1"
    templates = dict(aseg="{subject_id}/mri/aparc+aseg.mgz",
                     head="{subject_id}/mri/" + head_image + ".mgz")
    datasource = Node(SelectFiles(templates,
                                  base_directory=data_dir),
                      "datasource")

    # Convert images to nifti storage and float representation
    cvtaseg = Node(fs.MRIConvert(out_type="niigz"), "convertaseg")

    cvthead = Node(fs.MRIConvert(out_type="niigz",
                                 out_datatype="float"),
                   "converthead")

    # Turn the aparc+aseg into a brainmask
    makemask = Node(fs.Binarize(dilate=1, min=0.5), "makemask")

    # Extract the brain from the orig.mgz using the mask
    skullstrip = Node(fsl.ApplyMask(), "skullstrip")

    # FLIRT brain to MNI152_brain
    flirt = Node(fsl.FLIRT(reference=target_brain), "flirt")

    sw = [-180, 180]
    for dim in ["x", "y", "z"]:
        setattr(flirt.inputs, "searchr_%s" % dim, sw)

    # FNIRT head to MNI152
    fnirt = Node(fsl.FNIRT(ref_file=target_head,
                           refmask_file=target_mask,
                           config_file=fnirt_cfg,
                           fieldcoeff_file=True),
                 "fnirt")

    # Warp and rename the images
    warpbrain = Node(fsl.ApplyWarp(ref_file=target_head,
                                   interp="spline",
                                   out_file="brain_warp.nii.gz"),
                     "warpbrain")

    warpbrainhr = Node(fsl.ApplyWarp(ref_file=hires_head,
                                     interp="spline",
                                     out_file="brain_warp_hires.nii.gz"),
                       "warpbrainhr")

    # Generate a png summarizing the registration
    warpreport = Node(WarpReport(), "warpreport")

    # Save relevant files to the data directory
    fnirt_subs = [(head_image + "_out_masked_flirt.mat", "affine.mat"),
                  (head_image + "_out_fieldwarp", "warpfield"),
                  (head_image + "_out_masked", "brain"),
                  (head_image + "_out", "T1")]
    datasink = Node(DataSink(base_directory=data_dir,
                             parameterization=False,
                             substitutions=fnirt_subs),
                    "datasink")

    # Define and connect the workflow
    # -------------------------------

    normalize = Workflow(name=name)

    normalize.connect([
        (subjectsource, datasource,
            [("subject_id", "subject_id")]),
        (datasource, cvtaseg,
            [("aseg", "in_file")]),
        (datasource, cvthead,
            [("head", "in_file")]),
        (cvtaseg, makemask,
            [("out_file", "in_file")]),
        (cvthead, skullstrip,
            [("out_file", "in_file")]),
        (makemask, skullstrip,
            [("binary_file", "mask_file")]),
        (skullstrip, flirt,
            [("out_file", "in_file")]),
        (flirt, fnirt,
            [("out_matrix_file", "affine_file")]),
        (cvthead, fnirt,
            [("out_file", "in_file")]),
        (skullstrip, warpbrain,
            [("out_file", "in_file")]),
        (fnirt, warpbrain,
            [("fieldcoeff_file", "field_file")]),
        (skullstrip, warpbrainhr,
            [("out_file", "in_file")]),
        (fnirt, warpbrainhr,
            [("fieldcoeff_file", "field_file")]),
        (warpbrain, warpreport,
            [("out_file", "in_file")]),
        (subjectsource, datasink,
            [("subject_id", "container")]),
        (skullstrip, datasink,
            [("out_file", "normalization.@brain")]),
        (cvthead, datasink,
            [("out_file", "normalization.@t1")]),
        (flirt, datasink,
            [("out_file", "normalization.@brain_flirted")]),
        (flirt, datasink,
            [("out_matrix_file", "normalization.@affine")]),
        (warpbrain, datasink,
            [("out_file", "normalization.@brain_warped")]),
        (warpbrainhr, datasink,
            [("out_file", "normalization.@brain_hires")]),
        (fnirt, datasink,
            [("fieldcoeff_file", "normalization.@warpfield")]),
        (warpreport, datasink,
            [("out_file", "normalization.@report")]),
    ])

    return normalize


def create_ants_workflow(data_dir=None, subjects=None, name="antswarp"):
    """Set up the anatomical normalzation workflow using ANTS.

    Your anatomical data must have been processed in Freesurfer.
    Unlike most lyman workflows, the DataGrabber and DataSink
    nodes are hardwired within the returned workflow, as this
    tightly integrates with the Freesurfer subjects directory
    structure.

    Parameters
    ----------
    data_dir : path
        top level of data hierarchy/FS subjects directory
    subjects : list of strings
        list of subject IDs
    name : alphanumeric string, optional
        workflow name

    """
    if data_dir is None:
        data_dir = os.environ["SUBJECTS_DIR"]
    if subjects is None:
        subjects = []

    # Subject source node
    subjectsource = Node(IdentityInterface(fields=["subject_id"]),
                         iterables=("subject_id", subjects),
                         name="subjectsource")

    # Grab recon-all outputs
    templates = dict(aseg="{subject_id}/mri/aparc+aseg.mgz",
                     head="{subject_id}/mri/brain.mgz")
    datasource = Node(SelectFiles(templates,
                                  base_directory=data_dir),
                      "datasource")

    # Convert images to nifti storage and float representation
    cvtaseg = Node(fs.MRIConvert(out_type="niigz"), "convertaseg")

    cvtbrain = Node(fs.MRIConvert(out_type="niigz",
                                  out_datatype="float"),
                    "convertbrain")

    # Turn the aparc+aseg into a brainmask
    makemask = Node(fs.Binarize(dilate=4, erode=3, min=0.5), "makemask")

    # Extract the brain from the orig.mgz using the mask
    skullstrip = Node(fsl.ApplyMask(), "skullstrip")

    # Normalize using ANTS
    antswarp = Node(ANTSIntroduction(), "antswarp")

    # Generate a png summarizing the registration
    warpreport = Node(WarpReport(), "warpreport")

    # Save relevant files to the data directory
    datasink = Node(DataSink(base_directory=data_dir,
                             parameterization=False),
                    name="datasink")

    # Define and connect the workflow
    # -------------------------------

    normalize = Workflow(name=name)

    normalize.connect([
        (subjectsource, datasource,
            [("subject_id", "subject_id")]),
        (datasource, cvtaseg,
            [("aseg", "in_file")]),
        (datasource, cvtbrain,
            [("head", "in_file")]),
        (cvtaseg, makemask,
            [("out_file", "in_file")]),
        (cvtbrain, skullstrip,
            [("out_file", "in_file")]),
        (makemask, skullstrip,
            [("binary_file", "mask_file")]),
        (skullstrip, antswarp,
            [("out_file", "in_file")]),
        (antswarp, warpreport,
            [("brain_file", "in_file")]),
        (subjectsource, datasink,
            [("subject_id", "container")]),
        (antswarp, datasink,
            [("warp_file", "normalization.@warpfield"),
             ("inv_warp_file", "normalization.@inverse_warpfield"),
             ("affine_file", "normalization.@affine"),
             ("brain_file", "normalization.@brain")]),
        (warpreport, datasink,
            [("out_file", "normalization.@report")]),
    ])

    return normalize


class ANTSIntroductionInputSpec(BaseInterfaceInputSpec):

    in_file = File(exists=True)


class ANTSIntroductionOutputSpec(TraitedSpec):

    brain_file = File(exitsts=True)
    warp_file = File(exitsts=True)
    inv_warp_file = File(exitsts=True)
    affine_file = File(exitsts=True)


class ANTSIntroduction(BaseInterface):

    input_spec = ANTSIntroductionInputSpec
    output_spec = ANTSIntroductionOutputSpec

    def _run_interface(self, runtime):

        target_brain = fsl.Info.standard_image("avg152T1_brain.nii.gz")
        cmdline = " ".join(["antsIntroduction.sh",
                            "-d", "3",
                            "-r", target_brain,
                            "-i", self.inputs.in_file,
                            "-o", "ants_"])

        runtime.environ["ITK_NUM_THREADS"] = "1"
        runtime = submit_cmdline(runtime, cmdline)

        os.rename("ants_Affine.txt", "affine.txt")
        os.rename("ants_Warp.nii.gz", "warpfield.nii.gz")
        os.rename("ants_deformed.nii.gz", "brain_warp.nii.gz")
        os.rename("ants_InverseWarp.nii.gz", "inverse_warpfield.nii.gz")

        return runtime

    def _list_outputs(self):

        outputs = self._outputs().get()
        outputs["affine_file"] = op.abspath("affine.txt")
        outputs["warp_file"] = op.abspath("warpfield.nii.gz")
        outputs["brain_file"] = op.abspath("brain_warp.nii.gz")
        outputs["inv_warp_file"] = op.abspath("inverse_warpfield.nii.gz")

        return outputs


class WarpReportInput(BaseInterfaceInputSpec):

    in_file = File(exists=True)


class WarpReportOutput(TraitedSpec):

    out_file = File(exitsts=True)


class WarpReport(BaseInterface):

    input_spec = WarpReportInput
    output_spec = WarpReportOutput

    def _run_interface(self, runtime):

        target_brain = fsl.Info.standard_image("avg152T1_brain.nii.gz")
        m = Mosaic(self.inputs.in_file, target_brain)
        m.plot_contours("Reds", 2)
        m.savefig("warp_report.png")
        m.close()

        return runtime

    def _list_outputs(self):

        outputs = self._outputs().get()
        outputs["out_file"] = op.realpath("warp_report.png")
        return outputs



import inspect
import numpy as np
import scipy as sp
import pandas as pd
from scipy import stats
from sklearn.naive_bayes import GaussianNB
from sklearn.cross_validation import (cross_val_score,
                                      LeaveOneOut,
                                      LeaveOneLabelOut,
                                      KFold)

from numpy.testing import assert_array_equal, assert_array_almost_equal
import numpy.testing as npt
import nose.tools
from nose.tools import assert_equal, raises

from .. import mvpa

evs = [np.array([[6, 0, 1],
                 [18, 0, 1]]),
       np.array([[12, 0, 1],
                 [24, 0, 1]])]

dataset = dict(X=stats.norm(0, 1).rvs((24, 12)),
               y=stats.bernoulli(.5).rvs(24),
               runs=np.repeat([0, 1], 12))

dataset_3d = dict(X=stats.norm(0, 1).rvs((4, 24, 12)),
                  y=stats.bernoulli(.5).rvs(24),
                  runs=np.repeat([0, 1], 12))


def test_extract_dataset():
    """Test simple case."""
    evs = pd.DataFrame(dict(onset=[1, 2, 3],
                            condition=["foo", "foo", "bar"]),
                       dtype=float)
    ts = np.random.randn(5, 5, 5, 4)
    mask = ts[..., 0] > .5
    X, y, m = mvpa.extract_dataset(evs, ts, mask, 1)

    assert_array_equal(y, np.array([1, 1, 0]))

    should_be = sp.stats.zscore(ts[mask].T[np.array([1, 2, 3])])
    assert_array_equal(X, should_be)

    X_, y_, m_ = mvpa.extract_dataset(evs, ts, mask, 1,
                                      event_names=["bar", "foo"])
    assert_array_equal(X_, X)
    assert_array_equal(y_, y)
    assert_array_equal(m_, m)


def test_extract_sizes():
    """Test different frame sizes."""
    evs = pd.DataFrame(dict(onset=[1, 2, 3],
                            condition=["foo", "foo", "bar"]),
                       dtype=float)
    ts = np.random.randn(5, 5, 5, 4)
    mask = ts[..., 0] > .5

    X_1, y_1, m_1 = mvpa.extract_dataset(evs, ts, mask, 1)
    assert_equal(X_1.shape, (3, mask.sum()))
    assert_equal(len(m_1), mask.sum())

    X_2, y_2, m_2 = mvpa.extract_dataset(evs, ts, mask, 1, [-1, 0])
    assert_equal(X_2.shape, (2, 3, mask.sum()))


def test_extract_upsample():
    """Test upsampling during extraction."""
    evs = pd.DataFrame(dict(onset=[1, 2, 3],
                            condition=["foo", "foo", "bar"]),
                       dtype=float)
    ts = np.random.randn(5, 5, 5, 10)
    mask = ts[..., 0] > .5

    X, y, m = mvpa.extract_dataset(evs, ts, mask, tr=1,
                                   frames=[-1, 0], upsample=2)
    assert_equal(X.shape, (4, 3, mask.sum()))
    assert_equal(len(m), mask.sum())


def test_extract_zero_var():
    """Test that zero-variance features are masked out."""
    evs = pd.DataFrame(dict(onset=[1, 2, 3],
                            condition=["foo", "foo", "bar"]),
                       dtype=float)
    ts = np.random.randn(5, 5, 5, 10)
    ts[0, 0, 0, :] = 0
    mask = np.ones((5, 5, 5), np.bool)

    X, y, m = mvpa.extract_dataset(evs, ts, mask, 1)
    assert_array_equal(m, np.array([0] + [1] * (5 ** 3 - 1), np.bool))


@raises(ValueError)
def test_extract_mask_error():
    """Make sure mask is enforced as boolean."""
    evs = pd.DataFrame(dict(onset=[1], condition="foo"), dtype=float)
    ts = np.random.randn(10, 10, 10, 5)
    mask = np.random.rand(10, 10, 10)
    mvpa.extract_dataset(evs, ts, mask)


def test_decode_shapes():
    """Test that we get expected shapes from decode function."""
    model = GaussianNB()
    accs = mvpa._decode_subject(dataset, model)
    assert_equal(accs.shape, (1,))
    accs = mvpa._decode_subject(dataset_3d, model)
    assert_equal(accs.shape, (4,))

    splits = stats.bernoulli(.5).rvs(24)
    accs = mvpa._decode_subject(dataset, model, split_pred=splits)
    assert_equal(accs.shape, (2,))
    accs = mvpa._decode_subject(dataset_3d, model, split_pred=splits)
    assert_equal(accs.shape, (4, 2))

    accs = mvpa._decode_subject(dataset, model, trialwise=True)
    assert_equal(accs.shape, (len(dataset["y"]),))
    accs = mvpa._decode_subject(dataset_3d, model, trialwise=True)
    assert_equal(accs.shape, (4, len(dataset["y"])))


def test_decode_options():
    """Test some other options for the decode function."""
    model = GaussianNB()
    mvpa._decode_subject(dataset, model)
    mvpa._decode_subject(dataset_3d, model)
    splits = stats.bernoulli(.5).rvs(24)
    mvpa._decode_subject(dataset, model, cv="sample")
    mvpa._decode_subject(dataset, model, cv=5)
    mvpa._decode_subject(dataset, model, split_pred=splits)
    mvpa._decode_subject(dataset, model, trialwise=True)
    mvpa._decode_subject(dataset, model, logits=True)


@raises(ValueError)
def test_trialwise_split_exclusivity():
    """Test that we can't split predictions and get trialwise scores."""
    model = GaussianNB()
    splits = stats.bernoulli(.5).rvs(24)
    mvpa._decode_subject(dataset, model, split_pred=splits, trialwise=True)


def test_decode_cross_val():
    """Test that cv strings are correct."""
    model = GaussianNB()
    X = dataset["X"]
    y = dataset["y"]

    acc1 = mvpa._decode_subject(dataset, model, cv="run")
    cv = LeaveOneLabelOut(dataset["runs"])
    acc2 = cross_val_score(model, X, y, cv=cv).mean()
    assert_array_almost_equal(acc1, acc2)

    acc1 = mvpa._decode_subject(dataset, model, cv="sample")
    cv = LeaveOneOut(24)
    acc2 = cross_val_score(model, X, y, cv=cv).mean()
    assert_array_almost_equal(acc1, acc2)

    acc1 = mvpa._decode_subject(dataset, model, cv=4)
    cv = KFold(len(y), 4)
    acc2 = cross_val_score(model, X, y, cv=cv).mean()
    assert_array_almost_equal(acc1, acc2)


def test_accs_vs_logits():
    """Test that accs and logits give consisitent information."""
    model = GaussianNB()
    accs = mvpa._decode_subject(dataset, model, trialwise=True)
    logits = mvpa._decode_subject(dataset, model,
                                  logits=True, trialwise=True)
    logit_accs = np.where(logits >= 0, 1., 0.)
    assert_array_equal(accs, logit_accs)



import nose.tools as nt
from .. import fileutils as fu


class TestAddSuffix(object):

    def test_fname_only(self):

        fname = "buz.nii"
        nt.assert_equal(fu.add_suffix(fname, "bang"), "buz_bang.nii")

    def test_gzipped(self):

        fname = "buz.nii.gz"
        nt.assert_equal(fu.add_suffix(fname, "bang"), "buz_bang.nii.gz")

    def test_absolute_path(self):

        fname = "/foo/bar/buz.nii"
        nt.assert_equal(fu.add_suffix(fname, "bang"), "/foo/bar/buz_bang.nii")

    def test_relative_path(self):

        fname = "foo/bar/buz.nii"
        nt.assert_equal(fu.add_suffix(fname, "bang"), "foo/bar/buz_bang.nii")



#! /usr/bin/env python
"""
Group fMRI analysis frontend for Lyman ecosystem.

"""
import os
import sys
import time
import shutil
import os.path as op
from textwrap import dedent
import argparse

import matplotlib as mpl
mpl.use("Agg")

import nipype
from nipype import Node, MapNode, SelectFiles, DataSink, IdentityInterface

import lyman
import lyman.workflows as wf
from lyman import tools


def main(arglist):
    """Main function for workflow setup and execution."""
    args = parse_args(arglist)

    # Get and process specific information
    project = lyman.gather_project_info()
    exp = lyman.gather_experiment_info(args.experiment, args.altmodel)

    if args.experiment is None:
        args.experiment = project["default_exp"]

    if args.altmodel:
        exp_name = "-".join([args.experiment, args.altmodel])
    else:
        exp_name = args.experiment

    # Make sure some paths are set properly
    os.environ["SUBJECTS_DIR"] = project["data_dir"]

    # Set roots of output storage
    anal_dir_base = op.join(project["analysis_dir"], exp_name)
    work_dir_base = op.join(project["working_dir"], exp_name)
    nipype.config.set("execution", "crashdump_dir", project["crash_dir"])

    # Subject source (no iterables here)
    subject_list = lyman.determine_subjects(args.subjects)
    subj_source = Node(IdentityInterface(fields=["subject_id"]),
                       name="subj_source")
    subj_source.inputs.subject_id = subject_list

    # Set up the regressors and contrasts
    regressors = dict(group_mean=[1] * len(subject_list))
    contrasts = [["group_mean", "T", ["group_mean"], [1]]]

    # Subject level contrast source
    contrast_source = Node(IdentityInterface(fields=["l1_contrast"]),
                           iterables=("l1_contrast", exp["contrast_names"]),
                           name="contrast_source")

    # Group workflow
    space = args.regspace
    wf_name = "_".join([space, args.output])
    if space == "mni":
        mfx, mfx_input, mfx_output = wf.create_volume_mixedfx_workflow(
            wf_name, subject_list, regressors, contrasts, exp)
    else:
        mfx, mfx_input, mfx_output = wf.create_surface_ols_workflow(
            wf_name, subject_list, exp)

    # Mixed effects inputs
    ffxspace = "mni" if space == "mni" else "epi"
    ffxsmooth = "unsmoothed" if args.unsmoothed else "smoothed"
    mfx_base = op.join("{subject_id}/ffx/%s/%s/{l1_contrast}" % (ffxspace,
                                                                 ffxsmooth))
    templates = dict(copes=op.join(mfx_base, "cope1.nii.gz"))
    if space == "mni":
        templates.update(dict(
            varcopes=op.join(mfx_base, "varcope1.nii.gz"),
            dofs=op.join(mfx_base, "tdof_t1.nii.gz")))
    else:
        templates.update(dict(
            reg_file=op.join(anal_dir_base, "{subject_id}/preproc/run_1",
                             "func2anat_tkreg.dat")))

    # Workflow source node
    mfx_source = MapNode(SelectFiles(templates,
                                     base_directory=anal_dir_base,
                                     sort_filelist=True),
                         "subject_id",
                         "mfx_source")

    # Workflow input connections
    mfx.connect([
        (contrast_source, mfx_source,
            [("l1_contrast", "l1_contrast")]),
        (contrast_source, mfx_input,
            [("l1_contrast", "l1_contrast")]),
        (subj_source, mfx_source,
            [("subject_id", "subject_id")]),
        (mfx_source, mfx_input,
            [("copes", "copes")])
                 ]),
    if space == "mni":
        mfx.connect([
            (mfx_source, mfx_input,
                [("varcopes", "varcopes"),
                 ("dofs", "dofs")]),
                     ])
    else:
        mfx.connect([
            (mfx_source, mfx_input,
                [("reg_file", "reg_file")]),
            (subj_source, mfx_input,
                [("subject_id", "subject_id")])
                     ])

    # Mixed effects outputs
    mfx_sink = Node(DataSink(base_directory="/".join([anal_dir_base,
                                                      args.output,
                                                      space]),
                             substitutions=[("/stats", "/"),
                                            ("/_hemi_", "/"),
                                            ("_glm_results", "")],
                             parameterization=True),
                    name="mfx_sink")

    mfx_outwrap = tools.OutputWrapper(mfx, subj_source,
                                      mfx_sink, mfx_output)
    mfx_outwrap.sink_outputs()
    mfx_outwrap.set_mapnode_substitutions(1)
    mfx_outwrap.add_regexp_substitutions([
        (r"_l1_contrast_[-\w]*/", "/"),
        (r"_mni_hemi_[lr]h", "")
        ])
    mfx.connect(contrast_source, "l1_contrast",
                mfx_sink, "container")

    # Set a few last things
    mfx.base_dir = work_dir_base

    # Execute
    lyman.run_workflow(mfx, args=args)

    # Clean up
    if project["rm_working_dir"]:
        shutil.rmtree(project["working_dir"])


def parse_args(arglist):
    """Take an arglist and return an argparse Namespace."""
    help = dedent("""
    Perform a basic group analysis in lyman.

    This script currently only handles one-sample group mean tests on each of
    the fixed-effects contrasts. It is possible to run the group model in the
    volume or on the surface, although the actual model changes depending on
    this choice.

    The volume model runs FSL's FLAME mixed effects for hierarchical inference,
    which uses the lower-level variance estimates, and it applies standard
    GRF-based correction for multiple comparisons.  The details of the
    model-fitting procedure are set in the experiment file, along with the
    thresholds used for correction.

    The surface model uses a standard ordinary least squares fit and does
    correction with an approach based on a Monte Carlo simulation of the null
    distribution of cluster sizes for smoothed Gaussian data. Fortunately, the
    simulations are cached so this runs very quickly. Unfortunately, the cached
    simulations used a whole-brain search space, so this will be overly
    conservative for partial-brain acquisitions.

    Because of how GRF-based correction works, the thresholded volume images
    only have positive voxels. It is up to you to define "negative" versions of
    any contrasts where you are interested in relative deactivation.  The
    surface correction does not have this constraint, and the test sign is
    configurable in the experiment file (and will thus apply to all contrasts).

    By default the results are written under `group` next to the subject level
    data in the lyman analysis directory, although the output directory name
    can be changed.

    Examples
    --------

    Note that the parameter switches match any unique short version
    of the full parameter name.

    run_group.py

        With no arguments, this will process the default experiment with the
        subjects defined in $LYMAN_DIR/subjects.txt in the MNI space using the
        MultiProc plugin with 4 processes.

    run_group.py -s pilot_subjects -r fsaverage -o pilot -unsmoothed

        This will processes the subjects defined in a file at
        $LYMAN_DIR/pilot_subjects.txt as above but with the surface workflow.
        Unsmoothed fixed effects parameter estimates will be sampled to the
        surface and smoothed there. The resulting files will be stored under
        <analysis_dir>/<experiment>/pilot/fsaverage/<contrast>/<hemi>

    run_group.py -e nback -a parametric -p sge -q batch.q

        This will process an alternate model for the `nback` experiment using
        the SGE plugin by submitting jobs to the batch.q queue.

    Usage Details
    -------------

    """)

    parser = tools.parser
    parser.description = help
    parser.formatter_class = argparse.RawDescriptionHelpFormatter
    parser.add_argument("-experiment", help="experimental paradigm")
    parser.add_argument("-altmodel", help="alternate model to fit")
    parser.add_argument("-regspace", default="mni",
                        choices=["mni", "fsaverage"],
                        help="common space for group analysis")
    parser.add_argument("-unsmoothed", action="store_true",
                        help="used unsmoothed fixed effects outputs")
    parser.add_argument("-output", default="group",
                        help="output directory name")
    return parser.parse_args(arglist)

if __name__ == "__main__":
    main(sys.argv[1:])



#! /usr/bin/env python
"""Make static images of lyman results using PySurfer."""
import os
import os.path as op
import sys
import tempfile
import argparse
from textwrap import dedent
from time import sleep

import numpy as np
from scipy import stats
import nibabel as nib
import matplotlib.image as mplimg
from mayavi import mlab
from surfer import Brain, io

import moss
import lyman


def main(arglist):

    # Parse the command line
    args = parse_args(arglist)

    # Configure mayavi
    mlab.options.offscreen = args.nowindow

    # Load the lyman data
    subjects = lyman.determine_subjects(args.subjects)
    project = lyman.gather_project_info()
    exp = lyman.gather_experiment_info(args.experiment, args.altmodel)
    contrasts = exp["contrast_names"]
    z_thresh = exp["cluster_zthresh"]

    # Get the full correct name for the experiment
    if args.experiment is None:
        exp_name = project["default_exp"]
    else:
        exp_name = args.experiment
    exp_base = exp_name
    if args.altmodel is not None:
        exp_name = "-".join([exp_base, args.altmodel])

    # Group-level
    # ===========

    if args.level == "group":
        temp_base = op.join(project["analysis_dir"], exp_name, args.output,
                            args.regspace, "{contrast}")
        if args.regspace == "fsaverage":
            sig_thresh = -np.log10(stats.norm.sf(z_thresh))
            sig_thresh = np.round(sig_thresh) * 10
            corr_sign = exp["surf_corr_sign"]
            sig_name = "cache.th%d.%s.sig.masked.mgh" % (sig_thresh, corr_sign)
            stat_temp = op.join(temp_base, "{hemi}/osgm", sig_name)
            mask_temp = op.join(temp_base, "{hemi}/mask.mgh")
            png_temp = op.join(temp_base, "{hemi}/osgm/zstat_threshold.png")
        else:
            stat_temp = op.join(temp_base, "{hemi}.zstat1_threshold.mgz")
            mask_temp = op.join(temp_base, "{hemi}.group_mask.mgz")
            png_temp = op.join(temp_base, "zstat1_threshold_surf.png")
            corr_sign = "pos"

        contrast_loop("fsaverage", contrasts, stat_temp, mask_temp, png_temp,
                      args, z_thresh, corr_sign)

    # Subject-level
    # =============

    elif args.level == "subject":
        temp_base = op.join(project["analysis_dir"], exp_name, "{subj}",
                            "ffx", args.regspace, "smoothed/{contrast}")
        mask_temp = op.join(temp_base, "{hemi}.mask.mgz")
        stat_temp = op.join(temp_base, "{hemi}.zstat1.mgz")
        png_temp = op.join(temp_base, "zstat1_surf.png")

        for subj in subjects:
            contrast_loop(subj, contrasts, stat_temp, mask_temp, png_temp,
                          args, 1.96, "abs")


def contrast_loop(subj, contrasts, stat_temp, mask_temp, png_temp,
                  args, z_thresh, sign):
    """Iterate over contrasts and make surface images."""
    config_opts = dict(background="white", width=600, height=370)

    for contrast in contrasts:

        # Calculate where the overlay should saturate
        z_max = calculate_sat_point(stat_temp, contrast, subj)
        png_panes = []
        for hemi in ["lh", "rh"]:

            # Initialize the brain object
            b_subj = subj if args.regspace == "epi" else "fsaverage"
            b = Brain(b_subj, hemi, args.geometry,
                      config_opts=config_opts)

            # Plot the mask
            mask_file = mask_temp.format(contrast=contrast,
                                         hemi=hemi, subj=subj)
            add_mask_overlay(b, mask_file)

            # Plot the overlay
            stat_file = stat_temp.format(contrast=contrast,
                                         hemi=hemi, subj=subj)
            add_stat_overlay(b, stat_file, z_thresh, z_max, sign,
                             sig_to_z=args.regspace == "fsaverage")
            png_panes.append(save_view_panes(b, sign, hemi))

            # Maybe close the current figure
            if args.close:
                b.close()

        # Stitch the hemisphere pngs together and save
        full_png = np.concatenate(png_panes, axis=1)
        png_file = png_temp.format(contrast=contrast, hemi=hemi, subj=subj)
        mplimg.imsave(png_file, full_png)


def calculate_sat_point(template, contrast, subj=None):
    """Calculate the point at which the colormap should saturate."""
    data = []
    for hemi in ["lh", "rh"]:
        hemi_file = template.format(contrast=contrast, subj=subj, hemi=hemi)
        hemi_data = nib.load(hemi_file).get_data()
        data.append(hemi_data)
    data = np.concatenate(data)
    return max(3.71, moss.percentiles(data, 98))


def add_mask_overlay(b, mask_file):
    """Gray-out vertices outside of the common-space mask."""
    mask_data = nib.load(mask_file).get_data()

    # Plot the mask
    mask_data = np.logical_not(mask_data.astype(bool)).squeeze()
    if mask_data.any():
        b.add_data(mask_data, min=0, max=10, thresh=.5,
                   colormap="bone", alpha=.6, colorbar=False)


def add_stat_overlay(b, stat_file, thresh, max, sign, sig_to_z=False):
    """Plot a surface-encoded statistical overlay."""
    stat_data = nib.load(stat_file).get_data()

    # Possibly convert -log10(p) images to z stats
    if sig_to_z:
        stat_sign = np.sign(stat_data)
        p_data = 10 ** -np.abs(stat_data)
        z_data = stats.norm.ppf(p_data)
        z_data[np.sign(z_data) != stat_sign] *= -1
        stat_data = z_data

    # Plot the statistical data
    plot_stat = ((sign == "pos" and (stat_data > thresh).any())
                 or (sign == "neg" and (stat_data < -thresh).any())
                 or (sign == "abs" and (np.abs(stat_data) > thresh).any()))
    if plot_stat:
        stat_data = stat_data.squeeze()
        b.add_overlay(stat_data, min=thresh, max=max, sign=sign, name="stat")


def remove_stat_overlay(b):
    """Remove any overlays from the brain."""
    for name, overlay in b.overlays.items():
        overlay.remove()


def save_view_panes(b, sign, hemi):
    """Save lat, med and ven views, return a stacked image array."""
    views = ["lat", "med", "ven"]
    image_panes = []
    for view in views:

        # Handle the colorbar
        if "stat" in b.overlays:
            show_pos = hemi == "rh" and view == "ven"
            show_neg = hemi == "lh" and view == "ven"
            if sign in ("pos", "abs"):
                b.overlays["stat"].pos_bar.visible = show_pos
            if sign in ("neg", "abs"):
                b.overlays["stat"].neg_bar.visible = show_neg

        # Set the view and screenshot
        b.show_view(view, distance=330)
        sleep(0.2)  # pause for show_view
        image_panes.append(b.screenshot())

    # Stitch the images together and return the array
    full_image = np.concatenate(image_panes, axis=0)

    return full_image


def parse_args(arglist):

    help = dedent("""
    Plot the outputs of lyman analyses on a 3D surface mesh.

    This script uses PySurfer to generate surface images, which can provide
    considerably more information about the distribution of activation than
    volume-based images. Because the 3D rendering can be difficult to work
    with, the script is outside of the Nipype workflows that actually generate
    the results. Unfortunately, that means the script cannot be parallelized
    and does not cache its intermediate results.

    Images can be generated either at the group level or at the subject level,
    in which case the fixed-effects outputs are plotted. Currently, the
    statistics are plotted as Z statistics (even for Freesurfer results, which
    are stored as -log10[p]), and regions that were not included in the
    analysis mask are grayed out to represent their non-inclusion. For the
    group-level plots, some aspects of how the results are rendered onto the
    cortex can be controlled through parameters in the experiment file. Other
    parameters are available as command-line options.

    In an effort to allow flexibility for some particular issues that PySurfer
    can have, a few command-line options control what happens with the PySurfer
    window during the execution of the script. By default, a window is actually
    opened as the plot is drawn, but this can happen offscreen on systems for
    which this is possible. Additionally, on some platforms closing multiple
    PySurfer figures will cause a segfault, so the plots can be drawn into a
    single window for the whole script (unfortunately this may cause issues
    with the orientation of the scene lighting).

    It is important to emphasize that because this script must be executed
    separately from the processing workflows, it is possible for the static
    images to get out of sync with the actual results. It is up to the user
    to ensure that this does not transpire by always updating the snapshots
    when rerunning the workflows.

    Examples
    --------

    Note that the parameter switches match any unique short version
    of the full parameter name.

    surface_snapshots.py

        With no arguments, this will make snapshots for the default experiment
        at the group level in MNI space.

    surface_snapshots.py -r fsaverage -o pilot

        Make snapshots from the outputs of the surface workflow that are stored
        in <analysis_dir>/<experiment>/pilot/fsaverage. The -log10(p) maps that
        are written to Freesurfer will be converted to Z stats before plotting.

    surface_snapshots.py -l subject -e nback -a parametric -r epi

        Make snapshots of the fixed-effects model outputs on the native surface
        for an alternate model of the `nback` experiment for all subjects
        defined in the $LYMAN_DIR/subjects.txt file.

    surface_snapshots.py -s subj1 subj2 -r mni -l subject -g smoothwm

        Plot the default experiment fixed effects model outputs for subjects
        `subj1` and `subj2` in MNI space on the `smoothwm` surface of the
        fsaverage brain.

    surface_snapshots.py -k

        Same as the first example, but don't close the PySurfer window until the
        very end of the script. This can avoid a nasty segmentation fault issue
        on some systems.

    surface_snapshots.py -n

        Same as the first example, but don't show a PySurfer window while
        drawing. This is not guaranteed to work on all systems.

    Usage Details
    -------------

    """)

    parser = argparse.ArgumentParser(description=help)
    parser.formatter_class = argparse.RawDescriptionHelpFormatter
    parser.add_argument("-subjects", nargs="*",
                        help=("list of subject ids, name of file in lyman  "
                              "directory, or full path to text file with  "
                              "subject ids"))
    parser.add_argument("-experiment", help="experimental paradigm")
    parser.add_argument("-altmodel", help="alternate model to fit")
    parser.add_argument("-level", choices=["subject", "group"],
                        default="group",
                        help="analysis level to make images from")
    parser.add_argument("-regspace", default="mni",
                        choices=["mni", "fsaverage", "epi"],
                        help="common space where data are registered")
    parser.add_argument("-output", default="group",
                        help="group analysis output name")
    parser.add_argument("-geometry", default="inflated",
                        help="surface geometry for the rendering.")
    parser.add_argument("-nowindow", action="store_true",
                        help="plot offscreen (does not work on all platforms")
    parser.add_argument("-keepopen", dest="close", action="store_false",
                        help="do not close each figure after plotting")
    return parser.parse_args(arglist)


if __name__ == "__main__":
    main(sys.argv[1:])



from itertools import product
import numpy as np
import pandas as pd
from nipype.interfaces.base import Bunch

import nose.tools as nt
import numpy.testing as npt

from .. import model


class TestModelSummary(object):

    rs = np.random.RandomState(101)

    def test_dot_by_slice(self):

        n_x, n_y, n_z, n_t, n_pe = 10, 11, 12, 20, 5
        X = Bunch(design_matrix=pd.DataFrame(self.rs.randn(n_t, n_pe)))
        pes = self.rs.randn(n_x, n_y, n_z, n_pe)

        ms = model.ModelSummary()
        out = ms.dot_by_slice(X, pes)

        nt.assert_equal(out.shape, (n_x, n_y, n_z, n_t))

        for i, j, k in product(range(n_x), range(n_y), range(n_z)):
            pe_ijk = pes[i, j, k]
            yhat_ijk = X.design_matrix.dot(pe_ijk)
            npt.assert_array_almost_equal(yhat_ijk, out[i, j, k])

    def test_dot_by_slice_submatrix(self):

        n_x, n_y, n_z, n_t, n_pe = 11, 10, 9, 25, 4
        X = Bunch(design_matrix=pd.DataFrame(self.rs.randn(n_t, n_pe)),
                  confound_vector=np.array([[1, 0, 0, 1]]).T)
        pes = self.rs.randn(n_x, n_y, n_z, n_pe)

        ms = model.ModelSummary()
        out = ms.dot_by_slice(X, pes, "confound")

        nt.assert_equal(out.shape, (n_x, n_y, n_z, n_t))

        for i, j, k in product(range(n_x), range(n_y), range(n_z)):
            pe_ijk = pes[i, j, k] * X.confound_vector.T
            yhat_ijk = X.design_matrix.dot(pe_ijk.T).squeeze()
            npt.assert_array_almost_equal(yhat_ijk, out[i, j, k])

    def test_compute_r2(self):

        shape = 10, 10, 15, 20
        n_x, n_y, n_z, n_t = shape
        y = self.rs.randn(*shape)
        yhat = y + 2 * self.rs.randn(*shape)

        y -= y.mean(axis=-1)[..., np.newaxis]
        yhat -= yhat.mean(axis=-1)[..., np.newaxis]

        ms = model.ModelSummary()
        ms.y = y
        ms.sstot = np.sum(np.square(y), axis=-1)

        ssres, r2 = ms.compute_r2(yhat)
        for i, j, k in product(range(n_x), range(n_y), range(n_z)):
            y_ijk = y[i, j, k]
            yhat_ijk = yhat[i, j, k]
            ssres_ijk = np.sum(np.square(y_ijk - yhat_ijk))
            npt.assert_almost_equal(ssres[i, j, k], ssres_ijk)
            r2_ijk = 1 - ssres_ijk / np.square(y_ijk).sum()
            npt.assert_almost_equal(r2[i, j, k], r2_ijk)



import os
import os.path as op
import time
import tempfile
from nipype.interfaces.base import Bunch

import nose.tools as nt

from .. import submission


class TestSubmitCmdLine(object):

    def test_submit(self):

        tmp_dir = tempfile.gettempdir()
        tmp_file = op.join(tmp_dir, "tmp_{}".format(time.time()))
        cmdline = ["touch", tmp_file]
        runtime = Bunch(cwd=os.getcwd(), environ=os.environ)
        submission.submit_cmdline(runtime, cmdline)
        assert op.exists(tmp_file)
        os.remove(tmp_file)

    def test_stdout(self):

        cmdline = ["echo", "hello test"]
        runtime = Bunch(cwd=os.getcwd(), environ=os.environ)
        out = submission.submit_cmdline(runtime, cmdline)
        nt.assert_equal(out.stdout, "hello test\n")

    def test_stdout_addition(self):

        cmdline = ["echo", "oh why hello"]
        runtime = Bunch(stdout="hello test\n",
                        cwd=os.getcwd(), environ=os.environ)
        out = submission.submit_cmdline(runtime, cmdline)
        nt.assert_equal(out.stdout, "hello test\noh why hello\n")

    def test_runtime_error(self):

        tmp_dir = tempfile.gettempdir()
        tmp_file = op.join(tmp_dir, "i_am_not_a_file")
        cmdline = ["cat", tmp_file]
        runtime = Bunch(cwd=os.getcwd(), environ=os.environ)
        with nt.assert_raises(RuntimeError):
            out = submission.submit_cmdline(runtime, cmdline)
            error = "cat: {}: No such file or directory\n".format(tmp_file)
            nt.assert_equal(out.stderr, error)

    def test_cwd(self):

        cmdline = ["pwd"]
        home_dir = op.expanduser("~")
        runtime = Bunch(cwd=home_dir, environ=os.environ)
        out = submission.submit_cmdline(runtime, cmdline)
        nt.assert_equal(out.stdout, home_dir + "\n")



#! /usr/bin/env python
"""Create masks in native functional space from a variety of sources."""
import sys
import json
import time
import os.path as op
import argparse
from textwrap import dedent

from lyman.maskfactory import MaskFactory


def main(arglist):

    # Parse command line arguments
    args = parse_args(arglist)

    # Determine the type of processing we will do
    # First look for shortcut keys
    if args.label is not None:
        orig_type = "native_label" if args.native else "fsaverage_label"
    elif args.contrast is not None:
        orig_type = "stat_volume"
    elif args.aseg is not None:
        orig_type = "index_volume"
    # Otherwise we have to figure it out from the other arguments
    elif args.orig is not None and args.orig.endswith(".label"):
        orig_type = "native_label" if args.native else "fsaverage_label"
    elif args.id is not None:
        orig_type = "index_volume"
    else:
        raise ValueError("Could not determine orig type from arguments.")
    if args.debug:
        print "Processing type: %s" % orig_type

    # Initialise a factory object
    factory = MaskFactory(args.subjects, args.exp, args.roi, orig_type,
                          args.serial, args.debug)

    # Ensure that the orig file is an absolute path
    if args.orig is not None:
        args.orig = op.abspath(args.orig)

    # Execute the stream

    # Label files
    if "label" in orig_type:
        if args.label is not None:
            file_temp = "%(hemi)s." + args.label + ".label"
        hemis = ["lh", "rh"] if args.hemi is None else [args.hemi]
        if args.sample is not None:
            proj_args = dict(white=["frac", "0", "0", "0"],
                             graymid=["frac", ".5", ".5", "0"],
                             pial=["frac", "1", "1", "0"],
                             cortex=["frac", "0", "1", ".1"])[args.sample]
        else:
            proj_args = args.proj
        if orig_type == "native_label":
            label_temp = op.join(factory.data_dir, "%(subj)s",
                                 "label", file_temp)
            factory.from_native_label(label_temp, hemis, proj_args)
        elif orig_type == "fsaverage_label":
            label_temp = op.join(factory.data_dir, "fsaverage",
                                 "label", file_temp)
            factory.from_common_label(label_temp, hemis,
                                      proj_args, args.save_native)

    # Index volumes (atlas type)
    elif orig_type == "index_volume":
        if args.aseg:
            atlas_temp = op.join(factory.data_dir,
                                 "%(subj)s",
                                 "mri", "aseg.mgz")
        else:
            atlas_temp = args.orig

        factory.from_hires_atlas(atlas_temp, args.id, args.erode)

    # Combine mask image with a thresholded stat volume
    if args.contrast is not None:
        smooth_dir = "unsmoothed" if args.unsmoothed else "smoothed"
        if args.altmodel is None:
            exp = factory.experiment
        else:
            exp = factory.experiment + "-" + args.altmodel
        stat_file_temp = op.join(factory.anal_dir,
                                 exp,
                                 "%(subj)s",
                                 "ffx", "epi",
                                 smooth_dir,
                                 args.contrast,
                                 "zstat1.nii.gz")
        factory.apply_statistical_mask(stat_file_temp,
                                       args.thresh,
                                       args.nvoxels)

    # Write an image of the mask
    factory.write_png()

    # Get the current date
    args.created = time.asctime()
    # Write provenence information
    for subj in factory.subject_list:
        json_file = op.join(factory.data_dir, subj, "masks/%s.json" % args.roi)
        with open(json_file, "w") as fid:
            json.dump(args.__dict__, fid, sort_keys=True)


def parse_args(arglist):
    """Handle the command line."""
    help = dedent("""
    Create masks in native functional space from a variety of sources.

    Currently this can start with ROIs defined as a surface label on
    fsaverage, labels defined on each subject's native surface, ROIs
    defined on the high-res volume in Freesurfer space, or a statistical
    volume from a subject-level analysis.

    You can always pass a filepath (possibly with ``subj`` and ``hemi``
    string format keys to -orig and the program will work out what to
    do from the file type and other arguments. Alternatively, if files
    are in expected places (Freesurfer data hierarchy, lyman analysis
    hierarchy) there are shortcuts for the corresponding image type.

    The processing is almost entirely dependent on external binaries
    from FSL and Freesurfer, so both must be available.

    The resulting masks are defined in the space of the first functional
    run. This is also the target of the ``-regspace epi`` registration
    in the main lyman fmri workflows.

    The processing here is closely tied to these fmri workflows and requires
    subject-level preprocessing to have been performed. This program
    should be executed from a directory containing a project.py file
    that defines the relevant data and analysis paths.

    The script will also write a mosiac png with the mask overlaid on
    the mean functional image defining the epi space. This image can be
    viewed in ziegler. Additionally, it will write a json file with the command
    line argument dictionary for provenance tracking.

    If an IPython cluster is running, the processing will be executed
    in parallel by default on all available engines. This can be avoided
    by using the -serial option.

    Examples
    --------

    make_masks.py -s subj1 -roi ifs -label ifs -sample graymid

        Create a mask that will be saved to <data_dir>/subj1/masks/ifs.nii.gz
        that is defined on the common surface in bilateral files at
        <data_dir>/fsaverage/label/$hemi.ifs.label. When transforming from
        surface to volume space, project halfway from the white surface to
        the pial surface at each vertex and label all intersected voxels.

    make_masks.py -roi lh.ifs -label ifs -hemi lh -proj frac 0 1 .1

        Create a mask and save to <data_dir>/subj1/masks/lh.ifs.nii.gz
        that is defined on the common surface in the single, unilateral file
        <data_dir>/fsaverage/label/lh.ifs.label. This will create a mask for
        each subject defined in $LYMAN_DIR/subjects.txt. When transforming from
        surface to volume space, take samples in steps of 10% of the cortical
        thickness between the white and pial surfaces and label any intersected
        voxels.

    make_masks.py -roi V1 -orig labels/%(hemi)s.%(subj)_V1.label -native \ 
                  -sample graymid -s labeled_subjects

        Create masks from labels defined on native surfaces and stored outside
        the Freesurfer subjects directory hierarchy. Masks will be generated
        for all subjects defined in $LYMAN_DIR/labeled_subjects.txt.

    make_masks.py -roi ifs_hard -label ifs -contrast hard \ 
                  -thresh 2.3 -sample white

        Create a mask defined from the bilateral 'ifs' label as above that is
        intersected with a mask created by thresholding the fixed effect zstat
        for the 'hard' contrast of the default experiment at 2.3. This requires
        that registration and fixed effects have been run in the `epi` space.

    make_masks.py -roi wm -aseg -id 2 41 -erode 3

        Create a white matter mask from the aseg (Freesurfer auto-segmentation
        volume) for the default set of subjects. The hires mask will be eroded
        in 3 iterations before transformation to functional space.

    make_masks.py -roi caudate_hard -aseg -id 11 50 \ 
                  -exp stroop -contrast name_color -thresh 4

        Create a mask by intersecting activations in the 'name_color' contrast
        for the 'stroop' experiment with a caudate mask from the automatic
        segmentation.

    Usage Details
    -------------

    """)

    parser = argparse.ArgumentParser(description=help,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    # Necessary arguments
    parser.add_argument("-s", "-subjects", nargs="*", dest="subjects",
                        help="lyman subjects argument")
    parser.add_argument("-roi", required=True,
                        help="will form name of output mask file")

    # Optional argument (that always applies)
    parser.add_argument("-exp",
                        help="experiment (can use default from project.py)")

    # Generic input
    parser.add_argument("-orig",
        help="path to original file with subj and hemi format keys")

    # Label relavent arguments
    parser.add_argument("-label", help="label name if in Freesurfer hierachy")
    parser.add_argument("-native", action="store_true",
                        help="orig label is defined on native surface")
    parser.add_argument("-hemi", choices=["lh", "rh"],
        help="hemisphere if unilateral label - otherwise combine both")
    parser.add_argument("-sample",
                        choices=["white", "graymid", "pial", "cortex"],
                        help="shortcut for projection arguments")
    parser.add_argument("-proj", nargs=4,
                        help="projection args passed directly mri_label2vol")
    parser.add_argument("-save_native", action="store_true",
                        help="save label file after warping from common space")

    # Atlas-type image relevant images
    parser.add_argument("-aseg", action="store_true",
                        help="atlas image is aseg.mgz")
    parser.add_argument("-erode", type=int,
                        help="erode the hires volume with this many steps.")
    parser.add_argument("-id", type=int, nargs="*",
                        help="roi id(s) if orig is index volume")

    # Thresholded statistic relevant arguments
    parser.add_argument("-contrast",
                        help="first-level contrast to binarize z-stat map")
    parser.add_argument("-thresh", help="z-stat threshold")
    parser.add_argument("-nvoxels", type=int, help="take top <n> voxels from z-stat")
    parser.add_argument("-unsmoothed", action="store_true",
                        help="use unsmoothed fixed effects zstats")
    parser.add_argument("-altmodel",
                        help="stat file is from alternative model")

    # Generic execution relevant arguments
    parser.add_argument("-serial", action="store_true",
                        help="force serial execution")
    parser.add_argument("-debug", action="store_true",
                        help="enable debug mode")

    return parser.parse_args(arglist)


if __name__ == "__main__":
    main(sys.argv[1:])



#! /usr/bin/env python
#
# Copyright (C) 2012-2014 Michael Waskom <mwaskom@stanford.edu>

descr = """Lyman: A reproducible ecosystem for analyzing neuroimaging data."""

import os
from setuptools import setup

DISTNAME = 'lyman'
DESCRIPTION = descr
MAINTAINER = 'Michael Waskom'
MAINTAINER_EMAIL = 'mwaskom@stanford.edu'
LICENSE = 'BSD (3-clause)'
URL = 'http://stanford.edu/~mwaskom/software/lyman/'
DOWNLOAD_URL = 'https://github.com/mwaskom/lyman'
VERSION = '0.0.7'

def check_dependencies():

    # Just make sure dependencies exist, I haven't rigorously
    # tested what the minimal versions that will work are
    needed_deps = ["IPython", "numpy", "scipy", "matplotlib",
                   "sklearn", "skimage", "pandas", "statsmodels", 
                   "nibabel", "nipype", "seaborn"]
    missing_deps = []
    for dep in needed_deps:
        try:
            __import__(dep)
        except ImportError:
            missing_deps.append(dep)

    if missing_deps:
        missing = (", ".join(missing_deps)
                   .replace("sklearn", "scikit-learn")
                   .replace("skimage", "scikit-image"))
        raise ImportError("Missing dependencies: %s" % missing)

if __name__ == "__main__":

    if os.path.exists('MANIFEST'):
        os.remove('MANIFEST')

    import sys
    if not (len(sys.argv) >= 2 and ('--help' in sys.argv[1:] or
            sys.argv[1] in ('--help-commands',
                            '--version',
                            'egg_info',
                            'clean'))):
        check_dependencies()

    setup(name=DISTNAME,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        license=LICENSE,
        version=VERSION,
        url=URL,
        download_url=DOWNLOAD_URL,
        install_requires=["moss>=0.3.3"],
        packages=['lyman', 'lyman.tests',
                  'lyman.workflows', 'lyman.workflows.tests',
                  'lyman.tools', 'lyman.tools.tests'],
        scripts=['scripts/run_fmri.py', 'scripts/run_group.py',
                 'scripts/run_warp.py', 'scripts/setup_project.py',
                 'scripts/make_masks.py', 'scripts/anatomy_snapshots.py',
                 'scripts/surface_snapshots.py'],
        classifiers=[
                     'Intended Audience :: Science/Research',
                     'Programming Language :: Python :: 2.7',
                     'License :: OSI Approved :: BSD License',
                     'Operating System :: POSIX',
                     'Operating System :: Unix',
                     'Operating System :: MacOS'],
    )



from __future__ import division
import os
import os.path as op
from glob import glob
import hashlib
import re

import numpy as np
import scipy as sp
from scipy import stats
from scipy.interpolate import interp1d
import pandas as pd
import nibabel as nib
from sklearn.cross_validation import (KFold,
                                      LeaveOneOut,
                                      LeaveOneLabelOut)

import moss
from lyman import gather_project_info, gather_experiment_info


def extract_dataset(sched, timeseries, mask, tr=2, frames=None,
                    upsample=None, event_names=None):
    """Extract model and targets for single run of fMRI data.

    Parameters
    ----------
    sched : event sequence DataFrame
        must contain `condition` and `onsets` columns
    timeseries : 4D numpy array
        BOLD data
    mask : 3D boolean array
        ROI mask
    tr : int
        acquistion TR (in seconds)
    frames : sequence of ints, optional
        extract frames relative to event onsets or at onsets if None
    upsample : int
        upsample the raw timeseries by this factor using cubic spline
        interpolation
    event_names : list of strings
        list of condition names to use, otherwise uses sorted unique
        values in sched.condition

    Returns
    -------
    X : (n_frame) x n_samp x n_feat array
        model matrix (zscored by feature)
        if n_frame is 1, matrix is 2D
        NOTE: this array can have NaNs; use mask to remove
    y : n_ev vector
        target vector
    mask : n_feat boolean vector
        False where the variance for that feature is 0

    """
    # Set up the extraction frames
    if frames is None:
        frames = [0]
    elif not hasattr(frames, "__len__"):
        frames = [frames]
    frames = np.asarray(frames)

    if upsample is not None:
        n_frames = len(frames) * upsample
        frames = np.linspace(frames.min() * upsample,
                             (frames.max() + 1) * upsample,
                             n_frames + 1)[:-1]

    # Double check mask datatype
    if not mask.dtype == np.bool:
        raise ValueError("Mask must be boolean array")

    # Initialize the outputs
    if event_names is None:
        event_names = sorted(sched.condition.unique())
    else:
        sched = sched[sched.condition.isin(event_names)]
        event_names = list(event_names)
    X = np.zeros((len(frames), sched.shape[0], mask.sum()))
    y = sched.condition.map(lambda x: event_names.index(x))

    # Extract the ROI into a 2D n_tr x n_feat
    roi_data = timeseries[mask].T

    # Possibly upsample the raw data
    if upsample is None:
        upsample = 1
    else:
        time_points = len(roi_data)
        x = np.linspace(0, time_points - 1, time_points)
        xx = np.linspace(0, time_points,
                         time_points * upsample + 1)[:-upsample]
        interpolator = interp1d(x, roi_data, "cubic", axis=0)
        roi_data = interpolator(xx)

    # Build the data array
    for i, frame in enumerate(frames):
        onsets = np.array(sched.onset / tr).astype(int) * upsample
        onsets += int(frame)
        X[i, ...] = sp.stats.zscore(roi_data[onsets])

    # Find a mask to only use features with nonzero variance
    good_features = np.all([(np.var(X_i, axis=0) > 0) for X_i in X], axis=0)

    return X.squeeze(), y, good_features


def extract_subject(subj, problem, roi_name, mask_name=None, frames=None,
                    collapse=None, confounds=None, upsample=None,
                    smoothed=False, exp_name=None, event_names=None):
    """Build decoding dataset from predictable lyman outputs.

    This function will make use of the LYMAN_DIR environment variable
    to access information about where the relevant data live, so that
    must be set properly.

    This function caches its results and, on repeated calls,
    hashes the arguments and checks those against the hash value
    associated with the stored data. The hashing process considers
    the timestamp on the relevant data files, but not the data itself.

    Parameters
    ----------
    subj : string
        subject id
    problem : string
        problem name corresponding to design file name
    roi_name : string
        ROI name associated with data
    mask_name : string, optional
        name of ROI mask that can be found in data hierachy,
        uses roi_name if absent
    frames : int or sequence of ints, optional
        extract frames relative to event onsets or at onsets if None
    collapse : int, slice, or (subj x frames | frames) array
        if int, returns that element in first dimension
        if slice, take mean over the slice (both relative to
        frames, not to the actual onsets)
        if array, take weighted average of each frame (possibly
        with different weights by subject) otherwise return each frame
    confounds : string or list of strings
        column name(s) in schedule datafame to be regressed out of the
        data matrix during extraction
    upsample : int
        upsample the raw timeseries by this factor using cubic spline
        interpolation
    smoothed : bool
        whether to use the spatially smoothed timeseries data
    exp_name : string, optional
        lyman experiment name where timecourse data can be found
        in analysis hierarchy (uses default if None)
    event_names : list of strings
        list of condition names to use, otherwise uses sorted unique
        values in the condition field of the event schedule

    Returns
    -------
    data : dictionary
        dictionary with X, y, and runs entries, along with metadata

    """
    project = gather_project_info()
    exp = gather_experiment_info(exp_name)
    if exp_name is None:
        exp_name = project["default_exp"]

    if mask_name is None:
        mask_name = roi_name

    if smoothed:
        roi_name += "_smoothed"

    # Find the relevant disk location for the dataaset file
    ds_file = op.join(project["analysis_dir"],
                      exp_name, subj, "mvpa",
                      problem, roi_name, "dataset.npz")

    # Make sure the target location exists
    try:
        os.makedirs(op.dirname(ds_file))
    except OSError:
        pass

    # Get paths to the relevant files
    mask_file = op.join(project["data_dir"], subj, "masks",
                        "%s.nii.gz" % mask_name)
    design_file = op.join(project["data_dir"], subj, "design",
                          "%s.csv" % problem)
    smoothing = "smoothed" if smoothed else "unsmoothed"
    ts_dir = op.join(project["analysis_dir"], exp_name, subj,
                     "reg", "epi", smoothing)
    n_runs = len(glob(op.join(ts_dir, "run_*")))
    ts_files = [op.join(ts_dir, "run_%d/timeseries_xfm.nii.gz" % r_i)
                for r_i in range(1, n_runs + 1)]

    # Get the hash value for this dataset
    ds_hash = hashlib.sha1()
    ds_hash.update(mask_name)
    ds_hash.update(str(op.getmtime(mask_file)))
    ds_hash.update(str(op.getmtime(design_file)))
    for ts_file in ts_files:
        ds_hash.update(str(op.getmtime(ts_file)))
    ds_hash.update(np.asarray(frames).data)
    ds_hash.update(str(confounds))
    ds_hash.update(str(upsample))
    ds_hash.update(str(event_names))
    ds_hash = ds_hash.hexdigest()

    # If the file exists and the hash matches, convert to a dict and return
    if op.exists(ds_file):
        with np.load(ds_file) as ds_obj:
            if ds_hash == str(ds_obj["hash"]):
                dataset = dict(ds_obj.items())
                for k, v in dataset.items():
                    if v.dtype.kind == "S":
                        dataset[k] = str(v)
                # Possibly perform temporal compression
                _temporal_compression(collapse, dataset)
                return dataset

    # Otherwise, initialize outputs
    X, y, runs, use = [], [], [], []

    # Load mask file
    mask_data = nib.load(mask_file).get_data().astype(bool)

    # Load the event information
    sched = pd.read_csv(design_file)

    # Get a list of event names
    if event_names is None:
        event_names = sorted(sched.condition.unique())
    else:
        sched = sched[sched.condition.isin(event_names)]

    # Make each runs' dataset
    for r_i, sched_r in sched.groupby("run"):
        ts_data = nib.load(ts_files[int(r_i - 1)]).get_data()

        # Use the basic extractor function
        X_i, y_i, use_i = extract_dataset(sched_r, ts_data,
                                          mask_data, exp["TR"],
                                          frames, upsample, event_names)

        # Just add to list
        X.append(X_i)
        y.append(y_i)
        use.append(use_i)

    # Find the voxels that are good in every run and make a final mask
    good_features = np.all(use, axis=0)
    mask_data[mask_data] = good_features

    # Stick the list items together for final dataset
    if frames is not None and len(frames) > 1:
        X = np.concatenate(X, axis=1)
    else:
        X = np.concatenate(X, axis=0)
    y = np.concatenate(y)
    runs = sched.run

    # Apply the feature mask
    X = np.atleast_3d(X)[:, :, good_features].squeeze()

    # Regress the confound vector out from the data matrix
    if confounds is not None:
        X = np.atleast_3d(X)
        confounds = np.asarray(sched[confounds])
        confounds = stats.zscore(confounds.reshape(X.shape[1], -1))
        denom = confounds / np.dot(confounds.T, confounds)
        for X_i in X:
            X_i -= np.dot(X_i.T, confounds).T * denom
        X = X.squeeze()

    # Save to disk and return
    dataset = dict(X=X, y=y, runs=runs, roi_name=roi_name, subj=subj,
                   event_names=event_names, problem=problem, frames=frames,
                   confounds=confounds, upsample=upsample, smoothed=smoothed,
                   hash=ds_hash, mask=mask_data, mask_name=mask_name)
    np.savez(ds_file, **dataset)

    # Possibly perform temporal compression
    _temporal_compression(collapse, dataset)

    return dataset


def _temporal_compression(collapse, dset):
    """Either select a single frame or take the mean over several frames."""
    if collapse is not None:
        if isinstance(collapse, int):
            dset["X"] = dset["X"][collapse]
        elif isinstance(collapse, slice):
            dset["X"] = dset["X"][collapse].mean(axis=0)
        else:
            dset["X"] = np.average(dset["X"], axis=0, weights=collapse)

    dset["collapse"] = collapse


def extract_group(problem, roi_name, mask_name=None, frames=None,
                  collapse=None, confounds=None, upsample=None, smoothed=False,
                  exp_name=None, event_names=None, subjects=None, dv=None):
    """Load datasets for a group of subjects, possibly in parallel.

    Parameters
    ----------
    problem : string
        problem name corresponding to set of event types
    roi_name : string
        ROI name associated with data
    mask_name : string, optional
        name of ROI mask that can be found in data hierachy,
        uses roi_name if absent
    frames : int or sequence
        frames relative to stimulus onsets in event file to extract
    collapse : int, slice, or (subj x frames | frames) array
        if int, returns that element in first dimension
        if slice, take mean over the slice (both relative to
        frames, not to the actual onsets)
        if array, take weighted average of each frame (possibly
        with different weights by subject) otherwise return each frame
    confounds : sequence of arrays, optional
        list ofsubject-specific obs x n arrays of confounding variables
        to regress out of the data matrix during extraction
    upsample : int, optional
        upsample the raw timeseries by this factor using cubic splines
    smoothed : bool, optional
        whether to extract the spatially smoothed dataset
    exp_name : string, optional
        lyman experiment name where timecourse data can be found
        in analysis hierarchy (uses default if None)
    event_names : list of strings
        list of condition names to use, otherwise uses sorted unique
        values in the condition field of the event schedule
    subjects : sequence of strings, optional
        sequence of subjects to return; if none reads subjects.txt file
        from lyman directory and uses all defined there
    dv : IPython cluster direct view, optional
        if provided, executes in parallel using the cluster

    Returns
    -------
    data : list of dicts
       list of mvpa dictionaries

    """
    if subjects is None:
        subj_file = op.join(os.environ["LYMAN_DIR"], "subjects.txt")
        subjects = np.loadtxt(subj_file, str)
    subjects = list(subjects)

    # Allow to run in serial or parallel
    if dv is None:
        import __builtin__
        map = __builtin__.map
    else:
        map = dv.map_sync

    # Try to make frames a list, if possible
    if hasattr(frames, "tolist"):
        frames = frames.tolist()
    if hasattr(collapse, "tolist"):
        collapse = collapse.tolist()

    # Set up lists for the map to work
    problem = [problem for _ in subjects]
    roi_name = [roi_name for _ in subjects]
    mask_name = [mask_name for _ in subjects]
    frames = [frames for _ in subjects]
    if np.ndim(collapse) < 2:
        collapse = [collapse for _ in subjects]
    confounds = [confounds for _ in subjects]
    upsample = [upsample for _ in subjects]
    smoothed = [smoothed for _ in subjects]
    exp_name = [exp_name for _ in subjects]
    event_names = [event_names for _ in subjects]

    # Actually do the loading
    data = map(extract_subject, subjects, problem, roi_name, mask_name,
               frames, collapse, confounds, upsample, smoothed, exp_name,
               event_names)

    return data


def _results_fname(dataset, model, split_pred, trialwise, logits, shuffle,
                   exp_name):
    """Get a path to where files storing decoding results will live."""
    project = gather_project_info()
    if exp_name is None:
        exp_name = project["default_exp"]

    roi_name = dataset["roi_name"]
    collapse = dataset["collapse"]
    problem = dataset["problem"]
    subj = dataset["subj"]

    res_path = op.join(project["analysis_dir"],
                       exp_name, subj, "mvpa",
                       problem, roi_name)

    try:
        model_str = "_".join([i[0] for i in model.steps])
    except AttributeError:
        model_str = model.__class__.__name__

    collapse_str, split_str, trial_str, logit_str, shuffle_str = ("", "", "",
                                                                  "", "")
    if collapse is not None:
        if isinstance(collapse, slice):
            collapse_str = "%s-%s" % (collapse.start, collapse.stop)
        elif isinstance(collapse, int):
            collapse_str = str(collapse)
        else:
            collapse_str = "weighted"
    if split_pred is not None:
        split_str = "split"
        if hasattr(split_pred, "name"):
            if split_pred.name is None:
                split_str = split_pred.name
    if trialwise:
        trial_str = "trialwise"
    if logits:
        logit_str = "logits"
    if shuffle:
        shuffle_str = "shuffle"

    res_fname = "_".join([model_str, collapse_str, split_str,
                          trial_str, logit_str, shuffle_str])
    res_fname = re.sub("_{2,}", "_", res_fname)
    res_fname = res_fname.strip("_") + ".npz"
    res_fname = op.join(res_path, res_fname)

    return res_fname


def _hash_decoder(ds, model, split_pred=None, n_iter=None, random_seed=None):
    """Hash the inputs to a decoding analysis."""
    ds_hash = hashlib.sha1()
    try:
        ds_hash.update(ds["X"].data)
    except AttributeError:
        ds_hash.update(ds["X"].copy().data)
    ds_hash.update(ds["y"].data)
    ds_hash.update(ds["runs"])
    ds_hash.update(str(model))
    if split_pred is not None:
        ds_hash.update(np.array(split_pred).data)
    if n_iter is not None:
        ds_hash.update(str(n_iter))
    if random_seed is not None:
        ds_hash.update(str(random_seed))
    return ds_hash.hexdigest()


def _decode_subject(dataset, model, cv="run", split_pred=None,
                    trialwise=False, logits=False):
    """Internal function for classification."""
    if split_pred is not None and trialwise:
        raise ValueError("Cannot use both `split_pred` and `trialwise`.")

    # Unpack the dataset
    X = dataset["X"]
    y = dataset["y"]
    runs = dataset["runs"]
    if X.ndim < 3:
        X = [X]

    # Set up the cross-validation
    if cv == "run":
        cv_ = LeaveOneLabelOut(runs)
    elif cv == "sample":
        cv_ = LeaveOneOut(len(y))
    elif isinstance(cv, int):
        cv_ = KFold(len(y), cv)
    else:
        raise ValueError("CV argument was not understood")

    # Cross-validate the model over frames
    scores = np.empty((len(X), len(y)))
    for i, X_i in enumerate(X):
        for train, test in cv_:
            if logits:
                ps = model.fit(X_i[train], y[train]).predict_proba(X_i[test])
                rows = np.arange(len(ps))
                ps = ps[rows, y[test]]
                ps = np.log(ps) - np.log(1 - ps)
            else:
                ps = model.fit(X_i[train], y[train]).predict(X_i[test])
                ps = ps == y[test]
            scores[i, test] = ps

    # Possibly bin by trial splits
    if split_pred is not None:
        n_bins = len(np.unique(split_pred))
        split_scores = np.empty((len(X), n_bins))
        for i, bin in enumerate(np.unique(split_pred)):
            split_scores[:, i] = scores[:, split_pred == bin].mean()
        scores = split_scores
    elif not trialwise:
        scores = scores.mean(axis=1)

    return np.atleast_1d(scores.squeeze())


def decode_subject(dataset, model, cv="run", split_pred=None,
                   trialwise=False, logits=False, exp_name=None):
    """Perform decoding on a single dataset.

    This function hashes the relevant inputs and uses that to store
    persistant data over multiple executions.

    Parameters
    ----------
    dataset : dict
        decoding dataset
    model : scikit-learn estimator
        model to decode with
    cv : "run" | "sample" | k
        cross validate over runs, over samples (leave-one-out), or
        over `k` folds.
    spit_pred : pandas series or array
        bin prediction accuracies by the index values in the array.
        note: cannot be used with `trialwise`.
    trialwise : bool, optional
        if False, return accuracy/logit on each trial; otherwise take
        mean, possibly over frame. note: cannot be used with `split_pred`.
    logits : bool, optional
        return continuous target logit value for each observation
    exp_name : string
        name of experiment, if not default

    Return
    ------
    scores : array
        squeezed array of scores with shape (n_frames, (n_splits | n_trials))

    """
    # Ensure some inputs
    if split_pred is not None:
        split_pred = np.asarray(split_pred)

    # Get a path to the results will live
    res_file = _results_fname(dataset, model, split_pred, trialwise,
                              logits, False, exp_name)

    # Hash the inputs to the decoder
    decoder_hash = _hash_decoder(dataset, model, split_pred)

    # If the file exists and the hash matches, load and return
    if op.exists(res_file):
        with np.load(res_file) as res_obj:
            if decoder_hash == str(res_obj["hash"]):
                return res_obj["scores"]

    # Do the decoding with a private function so we can test it
    # without dealing with all the persistance stuff
    scores = _decode_subject(dataset, model, cv, split_pred, trialwise, logits)

    # Save the scores to disk
    res_dict = dict(scores=scores, hash=decoder_hash)
    try:
        os.makedirs(op.dirname(res_file))
    except OSError:
        pass
    np.savez(res_file, **res_dict)

    return scores


def decode_group(datasets, model, cv="run", split_pred=None,
                 trialwise=False, logits=False, exp_name=None, dv=None):
    """Perform decoding on a sequence of datasets.

    Parameters
    ----------
    datasets : sequence of dicts
        one dataset per subject
    model : scikit-learn estimator
        model to decode with
    cv : "run" | "sample" | k
        cross validate over runs, over samples (leave-one-out), or
        over `k` folds.
    spit_pred : series/array or sequence of seires/arrays, optional
        bin prediction accuracies by the index values in the array.
        can pass one array to use for all datasets, or a list
        of arrays with the same length as the dataset list.
        splits will form last axis of returned accuracy array.
        note: cannot be used with `trialwise`.
    trialwise : bool, optional
        if False, return accuracy/logit on each trial; otherwise take
        mean, possibly over frame. note: cannot be used with `split_pred`.
    logits : bool, optional
        return continuous target logit value for each observation
    exp_name : string
        name of experiment, if not default
    dv : IPython cluster direct view, optional
        IPython cluster to decode in parallel

    Return
    ------
    all_scores : array
        array with possible dimensions in (subj, frame, split, trial)

    """
    if dv is None:
        import __builtin__
        map = __builtin__.map
    else:
        map = dv.map_sync

    # Set up the lists for the map
    model = [model for d in datasets]
    cv = [cv for _ in datasets]
    if split_pred is None or not np.iterable(split_pred[0]):
        split_pred = [split_pred for _ in datasets]
    trialwise = [trialwise for _ in datasets]
    logits = [logits for _ in datasets]
    exp_name = [exp_name for _ in datasets]

    # Do the decoding
    all_scores = map(decode_subject, datasets, model, cv, split_pred,
                     trialwise, logits, exp_name)

    return np.array(all_scores)


def classifier_permutations(datasets, model, n_iter=1000, cv_method="run",
                            random_seed=None, exp_name=None, dv=None):
    """Do a randomization test on a set of classifiers with cached results.

    The randomizations can be distributed over an IPython cluster using
    the ``dv`` argument. Note that unlike the decode_group function,
    the parallelization occurs within, rather than over subjects.

    Parameters
    ----------
    datasets : list of dictionaries
        each item in the list is an mvpa dictionary
    n_iter : int
        number of permutation iterations
    cv_method : run | sample | cv arg for cross_val_score
        cross validate over runs, over samples (leave-one-out)
        or otherwise something that can be provided to the cv
        argument for sklearn.cross_val_score
    random_state : int
        seed for random state to obtain stable permutations
    exp_name : string
        experiment name if not default
    dv : IPython direct view
        view onto IPython cluster for parallel execution over iterations

    Returns
    -------
    group_scores : list of arrays
        permutation array for each item in datasets

    """
    group_scores = []
    for i_s, data in enumerate(datasets):

        # Get a path to the results will live
        res_file = _results_fname(data, model, None, False,
                                  False, True, exp_name)

        # Hash the inputs to the decoder
        decoder_hash = _hash_decoder(data, model, n_iter=n_iter,
                                     random_seed=random_seed)

        # If the file exists and the hash matches, load and return
        if op.exists(res_file):
            with np.load(res_file) as res_obj:
                if decoder_hash == str(res_obj["hash"]):
                    group_scores.append(res_obj["scores"])
                    continue

        # Otherwise, do the test for this dataset
        p_vals, scores = moss.randomize_classifier(data, model, n_iter,
                                                   cv_method, random_seed,
                                                   return_dist=True, dv=dv)

        # Save the scores to disk
        res_dict = dict(scores=scores, hash=decoder_hash)
        np.savez(res_file, **res_dict)

        group_scores.append(scores)

    return np.array(group_scores)


def model_coefs(datasets, model, flat=True, exp_name=None):
    """Fit a model on all data and save the learned model weights.

    This does not work for datasets with > 1 frames.

    Parameters
    ----------
    datasets : list of dicts
        group mvpa datasets
    model : scikit-learn estimator
        decoding model
    flat : bool
        if False return in original data space (with voxels outside
        mask represented as NaN. otherwise return straight from model
    exp_name : string or None
        experiment name, otherwise uses project default

    Returns
    -------
    out_coefs : list of arrays
        model coefficients; form is determined by `flat` parameter

    """
    project = gather_project_info()
    if exp_name is None:
        exp_name = project["default_exp"]

    out_coefs = []

    # Iterate through the datasets
    for dset in datasets:
        subj = dset["subj"]

        # Get a header and affine matrix for the EPI space
        smoothed = "smoothed" if dset["smoothed"] else "unsmoothed"
        epi_file = op.join(project["analysis_dir"], exp_name, subj,
                           "preproc/run_1/example_func.nii.gz")
        epi_img = nib.load(epi_file)
        epi_header, epi_affine = epi_img.get_header(), epi_img.get_affine()

        # Check if we need to do anything
        decoder_hash = _hash_decoder(dset, model, None)
        coef_file = _results_fname(dset, model, None, False,
                                   False, False, exp_name)
        coef_file = coef_file.strip(".npz") + "_coef.npz"
        coef_nifti = coef_file.strip(".npz") + ".nii.gz"
        if op.exists(coef_file) and op.exists(coef_nifti):
            with np.load(coef_file) as res_obj:
                if decoder_hash == str(res_obj["hash"]):
                    if flat:
                        data = res_obj["data"]
                    else:
                        data = nib.load(coef_nifti).get_data()
                    out_coefs.append(data)
                    continue

        # Determine the mask
        mask = dset["mask"]

        # Get the mask dimensions
        x, y, z = mask.shape

        # Fit the model and extract the learned model weights
        model = model.fit(dset["X"], dset["y"])
        if hasattr(model, "estimators_"):
            coef = np.array([e.coef_.ravel() for e in model.estimators_])
        else:
            coef = model.coef_
        coef_data = np.zeros((x, y, z, len(coef))) * np.nan
        coef_data[mask] = coef.T

        # Save the data both as a npz and nifti
        coef_dict = dict(data=coef, hash=decoder_hash)
        np.savez(coef_file, **coef_dict)
        coef_nifti = coef_file.strip(".npz") + ".nii.gz"
        coef_img = nib.Nifti1Image(coef_data, epi_affine, epi_header)
        nib.save(coef_img, coef_nifti)

    return out_coefs



#! /usr/bin/env python
"""
Script to guide project setup.

The basic idea here is borrowed from the Sphinx project, and
in fact a significant amount of the code in this file was
borrowed from there too.

"""
import sys, os, time, re
import os.path as op

_ansi_re = re.compile('\x1b\\[(\\d\\d;){0,2}\\d\\dm')
codes = {}

def get_terminal_width():
    """Borrowed from the py lib."""
    try:
        import termios, fcntl, struct
        call = fcntl.ioctl(0, termios.TIOCGWINSZ,
                           struct.pack('hhhh', 0, 0, 0, 0))
        height, width = struct.unpack('hhhh', call)[:2]
        terminal_width = width
    except (SystemExit, KeyboardInterrupt):
        raise
    except:
        # FALLBACK
        terminal_width = int(os.environ.get('COLUMNS', 80)) - 1
    return terminal_width


_tw = get_terminal_width()
def term_width_line(text):
    if not codes:
        # if no coloring, don't output fancy backspaces
        return text + '\n'
    else:
        # codes are not displayed, this must be taken into account
        return text.ljust(_tw + len(text) - len(_ansi_re.sub('', text))) + '\r'

def color_terminal():
    if not hasattr(sys.stdout, 'isatty'):
        return False
    if not sys.stdout.isatty():
        return False
    if 'COLORTERM' in os.environ:
        return True
    term = os.environ.get('TERM', 'dumb').lower()
    if term in ('xterm', 'linux') or 'color' in term:
        return True
    return False


def nocolor():
    codes.clear()

def coloron():
    codes.update(_orig_codes)

def colorize(name, text):
    return codes.get(name, '') + text + codes.get('reset', '')

def create_color_func(name):
    def inner(text):
        return colorize(name, text)
    globals()[name] = inner

_attrs = {
    'reset':     '39;49;00m',
    'bold':      '01m',
    'faint':     '02m',
    'standout':  '03m',
    'underline': '04m',
    'blink':     '05m',
}

for _name, _value in _attrs.items():
    codes[_name] = '\x1b[' + _value

_colors = [
    ('black',     'darkgray'),
    ('darkred',   'red'),
    ('darkgreen', 'green'),
    ('brown',     'yellow'),
    ('darkblue',  'blue'),
    ('purple',    'fuchsia'),
    ('turquoise', 'teal'),
    ('lightgray', 'white'),
]

for i, (dark, light) in enumerate(_colors):
    codes[dark] = '\x1b[%im' % (i+30)
    codes[light] = '\x1b[%i;01m' % (i+30)

_orig_codes = codes.copy()

for _name in codes:
    create_color_func(_name)

TERM_ENCODING = getattr(sys.stdin, 'encoding', None)

PROMPT_PREFIX = "> "

PROJECT_CONF = """\
# -*- coding: utf-8 -*-
#
# Configuration file for %(project_name)s built on %(now)s

# Default experiment name
default_exp = %(default_exp)s

# Data directory is where the input data lives
data_dir = '%(data_dir)s'

# Analysis directory is where anything written by these scripts will live
analysis_dir = '%(analysis_dir)s'

# Working directory is where data lives during workflow execution
working_dir = '%(working_dir)s'

# Crash directory is where debugging info will be written if things go wrong
crash_dir = '%(crash_dir)s'

# Set this to True to remove the working directory after each excecution
rm_working_dir = %(rm_work_dir)s

# Should normalization be performed using ANTS?
ants_normalization = %(ants_norm)s

"""

def mkdir_p(dir):
    if not op.isdir(dir):
        os.makedirs(dir)


class ValidationError(Exception):
    """Raised for validation errors."""

def is_path(x):
    if op.exists(x) and not op.isdir(x):
        raise ValidationError("Please enter a valid path name.")
    return x

def nonnull_string(s):
    if s is not None:
        return "'%s'" % s

def nonempty(x):
    if not x:
        raise ValidationError("Please enter some text.")
    return x

def choice(*l):
    def val(x):
        if x not in l:
            raise ValidationError('Please enter one of %s.' % ', '.join(l))
        return x
    return val

def boolean(x):
    if x.upper() not in ('Y', 'YES', 'N', 'NO'):
        raise ValidationError("Please enter either 'y' or 'n'.")
    return x.upper() in ('Y', 'YES')

def suffix(x):
    if not (x[0:1] == '.' and len(x) > 1):
        raise ValidationError("Please enter a file suffix, "
                              "e.g. '.rst' or '.txt'.")
    return x

def ok(x):
    return x


def do_prompt(d, key, text, default=None, validator=nonempty):
    while True:
        if default:
            prompt = blue(PROMPT_PREFIX + '%s [%s]: ' % (text, default))
        else:
            prompt = blue(PROMPT_PREFIX + text + ': ')
        x = raw_input(prompt)
        if default and not x:
            x = default
        if x.decode('ascii', 'replace').encode('ascii', 'replace') != x:
            if TERM_ENCODING:
                x = x.decode(TERM_ENCODING)
            else:
                print yellow('* Note: non-ASCII characters entered '
                             'and terminal encoding unknown -- assuming '
                             'UTF-8 or Latin-1.')
                try:
                    x = x.decode('utf-8')
                except UnicodeDecodeError:
                    x = x.decode('latin1')
        try:
            x = validator(x)
        except ValidationError, err:
            print red('* ' + str(err))
            continue
        break
    d[key] = x




def main():
    
    d = dict()

    if not color_terminal():
        nocolor()

    # Check if a project file already exists
    if op.exists("project.py"):

        # But let's make sure it's clean
        try:
            import project
            clean_import = True
        except Exception:
            clean_import = False
        import_notes = "" if clean_import else ", but it did not import cleanly"
        
        # Maybe give a heads up about it
        print red("Warning:"), """\
project.py file found in current directory%s.

Do you wish to generate a new project file?
(Note that you can always edit the existing file).
""" % import_notes

        # And let the user choose whether to overwrite it
        do_prompt(d, "overwrite", "Overwrite existing file? (y/N)",
                  "n", boolean)
        
        if not d["overwrite"]:
            print red("Aborting project setup.")
            sys.exit(0)
        os.remove("project.py")

    # Now go through the prompted setup procedure
    print bold("Let's set up your project.")

    print '''
Please enter values for the following settings (just press Enter to
accept a default value, if one is given in brackets).

Please use relative paths.
'''

    do_prompt(d, "project_name", "Project name")

    do_prompt(d, "default_exp", "Default experiment", None, nonnull_string)

    do_prompt(d, "data_dir", "Data tree path", "../data", is_path)

    do_prompt(d, "analysis_dir", "Analysis tree path", "../analysis", is_path)
    
    do_prompt(d, "working_dir", "Working tree path", 
              op.join(d['analysis_dir'], 'workingdir'), is_path)

    crash_stem = "niypype-" + os.environ.get("LOGNAME", "-") + "-crashes"
    do_prompt(d, "crash_dir", "Crashdump path",
              op.join("/tmp", crash_stem))

    do_prompt(d, "rm_work_dir", "Remove working directory after execution? (Y/n)",
              "y", boolean)

    do_prompt(d, "ants_norm", "Use ANTS for normalization? (y/N)",
              "n", boolean)


    # Record the time this happened
    d['now'] = time.asctime()

    # Write the project file
    f = open("project.py", "w")
    conf_text = PROJECT_CONF % d
    f.write(conf_text.encode("utf-8"))
    f.close()

if __name__ == "__main__":
    main()
